//
//  QAAppDelegate.m
//  EarthScene
//
//  Created by Daniel Dönigus on 12.09.12.
//  Copyright (c) 2012 Daniel Dönigus. All rights reserved.
//

#import "QAAppDelegate.h"
#import <OpenGL/OpenGL.h>

@implementation QAAppDelegate

#pragma mark - Loading textures manually

- (GLint) loadTextureWithName:(NSString*) filename andExtension:(NSString*) fileExtension {
    CFBundleRef gameBundle = CFBundleGetMainBundle();
    
    CFStringRef subdirectory = NULL;
    CFURLRef fileLocation;
    fileLocation = CFBundleCopyResourceURL(gameBundle, (__bridge CFStringRef)filename,
                                           (__bridge CFStringRef)fileExtension, subdirectory);
    
    CGImageSourceRef myImageSourceRef = CGImageSourceCreateWithURL (fileLocation, NULL);
    CGImageRef myImageRef = CGImageSourceCreateImageAtIndex (myImageSourceRef, 0, NULL);
    GLuint myTextureName;
    size_t width = CGImageGetWidth(myImageRef);
    size_t height = CGImageGetHeight(myImageRef);
    CGRect rect = {{0, 0}, {width, height}};
    void * myData = calloc(width * 4, height);
    CGColorSpaceRef space = CGColorSpaceCreateDeviceRGB();
    CGContextRef myBitmapContext = CGBitmapContextCreate (myData,
                                                          width, height, 8,
                                                          width*4, space,
                                                          kCGBitmapByteOrder32Host |
                                                          kCGImageAlphaPremultipliedFirst);
    CGContextSetBlendMode(myBitmapContext, kCGBlendModeCopy);
    CGContextDrawImage(myBitmapContext, rect, myImageRef);
    CGContextRelease(myBitmapContext);
    
    glPixelStorei(GL_UNPACK_ROW_LENGTH, (GLint)width);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glGenTextures(1, &myTextureName);
    glBindTexture(GL_TEXTURE_2D, myTextureName);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, (GLint)width, (GLint)height,
                 0, GL_BGRA_EXT, GL_UNSIGNED_INT_8_8_8_8_REV, myData);
    free(myData);
    
    return myTextureName;
}

#pragma mark - Setting up the scene and program at startup

- (void)retrieveEarthMaterial {
    //Retrieving the material of the earth geometry. This is later used to set the program for the own
    //shader when the user selected the menu item View->Own Shader
    SCNNode* rootNode = self.view.scene.rootNode;
    SCNNode* earthNode = [rootNode childNodeWithName:@"earth" recursively:NO];
    SCNGeometry* geometry = earthNode.geometry;
    self.earthMaterial = geometry.firstMaterial;
}

- (void)retrieveLightNode {
    //Retrieving the light node of the scene. This is later used to get the light position
    SCNNode* rootNode = self.view.scene.rootNode;
    self.lightNode = [rootNode childNodeWithName:@"light" recursively:NO];
}

- (void)retrieveMoonMaterial {
    //Retrieving the material of the moon geometry. This is later used to set the normal map when the
    //user selected the menu item View->Own Shader
    SCNNode* rootNode = self.view.scene.rootNode;
    SCNNode* moonNode = [rootNode childNodeWithName:@"moon" recursively:NO];
    self.moonMaterial = moonNode.geometry.firstMaterial;
    self.moonMaterial.normal.minificationFilter = SCNLinearFiltering;
    self.moonMaterial.normal.magnificationFilter = SCNLinearFiltering;
    self.moonMaterial.normal.mipFilter = SCNLinearFiltering;
}

- (void)loadMoonNormalMap {
    //Loading the moon normal map as NSImage. The image is set as material.diffuse.contents of the moon
    //geometry
    NSString * normalMapPath = [[NSBundle mainBundle]
                                pathForResource:@"moonnormal1k" ofType:@"jpg"];
    self.moonNormalMap = [[NSImage alloc] initWithContentsOfFile:normalMapPath];
}

- (void)createProgram {
    //Creating the program
    SCNProgram *program = [[SCNProgram alloc] init];
    //Loading vertex shader program from vertex.prog
    NSString * vertexPath = [[NSBundle mainBundle]
                           pathForResource:@"vertex" ofType:@"prog"];
    NSString *vertexShader = [NSString stringWithContentsOfFile:vertexPath encoding:NSUTF8StringEncoding error:nil];
    //Loading fragment shader program from vertex.prog
    NSString * fragmentPath = [[NSBundle mainBundle]
                               pathForResource:@"fragment" ofType:@"prog"];
    NSString *fragmentShader = [NSString stringWithContentsOfFile:fragmentPath encoding:NSUTF8StringEncoding error:nil];
    //Setting shader to program
    program.vertexShader = vertexShader;
    program.fragmentShader = fragmentShader;
    //The AppDelegate is also the delegate of our program
    program.delegate       = self;
    
    //Setting the semantics for our attributes in the vertex shader:
    //Vertex positions in geometry local space
    [program setSemantic:SCNGeometrySourceSemanticVertex forSymbol:@"a_Position" options:nil];
    //Normals of geometry
    [program setSemantic:SCNGeometrySourceSemanticNormal forSymbol:@"a_Normal" options:nil];
    //Texture coordinates of of geometry
    [program setSemantic:SCNGeometrySourceSemanticTexcoord forSymbol:@"a_TextureCoordinate" options:nil];

    //Setting the semantics for our uniforms in vertex and fragment shader:
    //Model-View-Projection Matrix
    [program setSemantic:SCNModelViewProjectionTransform forSymbol:@"u_ModelViewProjectionMatrix" options:nil];
    //Model-View Matrix
    [program setSemantic:SCNModelViewTransform forSymbol:@"u_ModelViewMatrix" options:nil];
    //View Matrix
    [program setSemantic:SCNViewTransform forSymbol:@"u_ViewMatrix" options:nil];
    //Normal Matrix
    [program setSemantic:SCNNormalTransform forSymbol:@"u_NormalMatrix" options:nil];
    
    self.earthProgram = program;
}

- (void) loadTextures {
    //Loading the textures for earth manually because the own shader needs them to be bound manually
    self.normalMap = [self loadTextureWithName:@"earthnormal1k" andExtension:@"jpg"];
    self.colorMap = [self loadTextureWithName:@"earthmap1k" andExtension:@"jpg"];
    self.specularMap = [self loadTextureWithName:@"earthspec1k" andExtension:@"jpg"];
    self.nightMap = [self loadTextureWithName:@"earthlights1k" andExtension:@"jpg"];
}

- (void) insertFloor {
    //Creating a new node that will contain the floor
    SCNNode* floorNode = [SCNNode node];
    floorNode.position = SCNVector3Make(0.0, 0.0, -3.0);
    //Creating floor geometry
    SCNFloor* floor = [SCNFloor floor];
    floor.reflectivity = 0.1;
    floorNode.geometry = floor;
    //Adding the floor to the scene graph
    [[self.view.scene rootNode] addChildNode:floorNode];
    
    //Getting the first material of the floor geometry
    SCNMaterial* material = floor.firstMaterial;
    //Setting the wood texture as diffuest content of floor material
    NSString * texturePath = [[NSBundle mainBundle] pathForResource:@"wood" ofType:@"jpg"];
    NSImage* image = [[NSImage alloc] initWithContentsOfFile:texturePath];
    material.diffuse.contents = image;
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    //Setting up scene and program
    [self retrieveEarthMaterial];
    [self retrieveLightNode];
    [self retrieveMoonMaterial];
    [self loadMoonNormalMap];
    [self createProgram];
    [self loadTextures];
    [self insertFloor];
}

#pragma mark - Program delegate

- (BOOL)program:(SCNProgram*)program bindValueForSymbol:(NSString*)symbol atLocation:(unsigned int)location programID:(unsigned int)programID renderer:(SCNRenderer*)renderer
{
    //This program delegate method is used to load additional informations to our shaders
    
    //Camera position
    if ([symbol isEqualToString:@"u_EyePosition"]) {
        SCNVector3 position = self.view.pointOfView.position;
        glUniform3f(location, position.x, position.y, position.z);
        return YES;
    }
    
    //Light position
    if ([symbol isEqualToString:@"u_LightPosition"]) {
        SCNVector3 position = self.lightNode.position;
        glUniform3f(location, position.x, position.y, position.z);
        return YES;
    }
    
    //Normal texture
    if ([symbol isEqualToString:@"u_TextureNormal"]) {
        glUniform1i(location, 0);
        glActiveTexture (GL_TEXTURE0);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, self.normalMap);
        return YES;
    }
    
    //Standard color texture
    if ([symbol isEqualToString:@"u_TextureColor"]) {
        glUniform1i(location, 1);
        glActiveTexture (GL_TEXTURE1);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, self.colorMap);
        return YES;
    }
    
    //Texture showing the illuminates cities
    if ([symbol isEqualToString:@"u_TextureNight"]) {
        glUniform1i(location, 2);
        glActiveTexture (GL_TEXTURE2);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, self.nightMap);
        return YES;
    }
    
    //Specular map, so only the sea reflects light
    if ([symbol isEqualToString:@"u_TextureSpecular"]) {
        glUniform1i(location, 3);
        glActiveTexture (GL_TEXTURE3);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, self.specularMap);
        return YES;
    }
    
    return NO;
}

- (void)program:(SCNProgram*)program unbindValueForSymbol:(NSString*)symbol atLocation:(unsigned int)location programID:(unsigned int)programID renderer:(SCNRenderer*)renderer
{
}


- (void)program:(SCNProgram*)program handleError:(NSError*)error
{
    NSLog(@"%@", [error localizedDescription]);
}

#pragma mark - Switching shaders

- (IBAction)changeShader:(id)sender {
    if (self.earthMaterial.program == nil) {
        //Set the own shader programs to be used to render the earth
        self.earthMaterial.program = self.earthProgram;
        //Setting the normal map for the moon geometry
        self.moonMaterial.normal.contents = self.moonNormalMap;
    } else {
        //Resetting the own shader programs and moon normal map
        self.earthMaterial.program = nil;
        self.moonMaterial.normal.contents = nil;
    }
}

@end
